
import collections
import warnings
from pandas import Series, DataFrame
import pandas as pd

from .read.dinosurfacelevel import DinoSurfaceLevel


class SwSeries:
    """ Surface water level series """

    LOCPROPS_NAMES = [
        'locname','alias','xcr','ycr',
        'level_reference','grid_reference']

    LEVELPROPS_NAMES = [
        'level',] #'remark',]

    REMARKPROPS_NAMES = [ 
        'remark']

    MAPPING_DINOLOCPROPS = collections.OrderedDict([
        ('locname','nitgcode'),
        ('alias','alias'),
        ('xcr','xcr'),
        ('ycr','ycr'),
        ('level_reference','reference'),
        ('grid_reference','grid_reference'),
        ])

    MAPPING_DINOLEVELPROPS = collections.OrderedDict([
        ('level','level'),])
        ##('remark','remark'),])

    MAPPING_DINOREMARKPROPS = collections.OrderedDict([
        ('remark','remark'),])


    def __repr__(self):

        name = self.name()
        nlev = len(self.levels())
        return (f'{name} (n={nlev})')


    def __init__(self,levels=None,locprops=None,remarks=None):

        if locprops is None:
            self._locprops = Series(index=self.LOCPROPS_NAMES)
        elif isinstance(locprops,pd.Series):
            self._locprops = locprops
        else:
            raise TypeError(f'Argument locprops is not a pandas Series but '
                f'{type(locprops)}')

        if levels is None:
            self._levels = DataFrame(columns=self.LEVELPROPS_NAMES)
        elif isinstance(levels,pd.DataFrame):
            self._levels = levels
        else:
            raise TypeError(f'Argument levels is not a pandas DataFrame '
                f'but {type(levels)}')

        if remarks is None:
            self._remarks = DataFrame(columns=self.REMARKPROPS_NAMES)
        elif isinstance(remarks,pd.DataFrame):
            self._remarks = remarks
        else:
            raise TypeError(f'Argument remarks is not a pandas DataFrame '
                f'but {type(remarks)}')


    @classmethod
    def from_dinocsv(cls,fpath):
        """Read dinoloket csv file with surface water level measurements
           and return data as SwSeries object"""

        dns = DinoSurfaceLevel(fpath)
        dinolevels = dns.levels()
        dinoprops = dns.locprops()
        dinoremarks = dns.remarks()

        locprops = Series(index=cls.LOCPROPS_NAMES)
        for propname in cls.LOCPROPS_NAMES:
            dnprop = cls.MAPPING_DINOLOCPROPS[propname]
            if dnprop in dinoprops:
                locprops[propname] = dinoprops[dnprop]
            else:
                msg = (f'{dnprop} not in dinocsv locprops names'
                    f'{list(dinoprops.index.values)}')
                warnings.warn(msg)

        levels = DataFrame(columns=cls.LEVELPROPS_NAMES)
        for propname in cls.LEVELPROPS_NAMES:
            dnprop = cls.MAPPING_DINOLEVELPROPS[propname]
            if dnprop in dinolevels:
                levels[propname] = dinolevels[dnprop]
            else:
                msg = (f'{dnprop} not in dinocsv level names'
                    f'{list(dinolevels.index.values)}')
                warnings.warn(msg)

        remarks = DataFrame(columns=cls.REMARKPROPS_NAMES)
        for propname in cls.REMARKPROPS_NAMES:
            dnprop = cls.MAPPING_DINOREMARKPROPS[propname]
            if dnprop in dinoremarks:
                remarks[propname] = dinoremarks[dnprop]
            else:
                msg = (f'{dnprop} not in dinocsv remarks'
                    f'{list(dinoremarks.index.values)}')
                warnings.warn(msg)

        return cls(levels=levels,locprops=locprops, remarks=remarks)


    def name(self):
        """Return water level gauge location name"""
        return self._locprops['locname']


    def levels(self,dropnan=True):
        """Return measured water levels

        Parameters
        ----------
        dropnan : boolean, default True
            drop dates with missing level values

        Returns
        -------
        pd.Series

        """

        if dropnan:
            sr = self._levels['level'].dropna()
        else:
            sr = self._levels['level']

        sr.name = self.name()
        return sr



    def stats(self):
        """Return descriptive statistics of surface water level series"""

        levels = self.levels(dropnan=True)
        sr = Series(dtype='object',name=self._locprops['locname'])

        ##sr['name'] = sr.name
        sr['nmeas'] = levels.count()
        sr['firstyear'] = levels.index.min().year
        sr['lastyear'] = levels.index.max().year

        sr['nyears'] = len(levels.groupby(levels.index.year).count())
        sr['yearspan'] = levels.index.year.max()-levels.index.year.min()+1

        sr['mean'] = levels.mean()
        sr['std'] = levels.std()
        sr['q05'] = levels.quantile(q=0.05)      
        sr['q50'] = levels.quantile(q=0.5)      
        sr['q95'] = levels.quantile(q=0.95)
        sr['q95-q05'] = levels.quantile(q=0.95)- levels.quantile(q=0.05)

        sr['ref'] = self._locprops['level_reference']
        sr['firstdate'] = levels.index.min()
        sr['lastdate'] = levels.index.max()
        sr['alias'] = self._locprops['alias']
        sr['xcr'] = self._locprops['xcr']
        sr['ycr'] = self._locprops['ycr']

        return sr

    def remarks(self,dropnan=True,locname=True):
        """Return remarks

        Parameters
        ----------
        dropnan : bool, default True
            drop rows with empty remarks
        locname : bool, default True
            add column with location name to all rows

        Return
        ------
        pd.Dataframe

        """

        rem = self._remarks.copy()
        rem.insert(0,'date',rem.index)
        rem = rem.reset_index(drop=True)

        if locname:
            rem.insert(0,'location',self.name())

        return rem[rem['remark']!='']





from setuptools import setup, find_packages

with open("README.MD", "r") as fh:
    long_description = fh.read()

setup(
    name='acequia',
    version='0.1.5',
    license='MIT',
    author='T.J. de Meij',
    author_email='thomasdemeij@gmail.com',
    description='Python package for reading, writing and analyzing'
                'groundwater head time series.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/acequia-package/acequia.git',
    install_requires=[
        'numpy',
        'matplotlib',
        'scipy',
        'pandas',
        'geopandas',
        'shapely',
        'seaborn',
        'statsmodels',
        'simplekml',
        'sphinx',
        ],

    # pip will copy non-code files when installing
    include_package_data=True,
    
    packages=find_packages(exclude=[]),
)




from pathlib import Path
import warnings
import numpy as np
import pandas as pd
from pandas import DataFrame, Series


class KnmiRain:
    """Read precipitation data from knmi file"""

    NAMESTR = 'STN,YYYYMMDD,   RD,   SX,' + 'empty'
    ##COLNAMES = ['STN','YYYYMMDD','RD','SX']
    COLNAMES = NAMESTR.replace(' ','').split(',')
    KEEPCOLS = COLNAMES
    VARNAMES = ['prc','snow']
    SNOWCODE = {
        '997': 'gebroken sneeuwdek/broken snow cover < 1 cm',
        '998': 'gebroken sneeuwdek/broken snow cover >=1 cm',
        '999': 'sneeuwhopen/snow dunes}',
        }
    SNOWVALS = {'997':'1','998':'1','999':'1'}
    SKIPROWS = 24


    def __init__(self,filepath=None):

        # NAMESTR[:-1] : trailing colon creates extra column
        self.fpath = Path(filepath)

        if not (self.fpath.exists() and self.fpath.is_file()):
            warnings.warn((f"Filepath '{self.fpath}' not found"
                'empty dataframe is returned.'))
            self.rawdata = pd.DataFrame(columns=self.COLNAMES)
            self.data = pd.DataFrame(columns=self.KEEPCOLS)
        else:
            self.rawdata = self._readfile()
            self.data = self._clean_rawdata(self.rawdata)


    def __repr__(self):
        return (f'{self.__class__.__name__} (n={len(self.data)})')


    def _readfile(self,fpath):

        # read csv to pd.DataFrame with only str values
        rawdata = pd.read_csv(fpath,sep=',',
            skiprows=self.SKIPROWS,
            names=self.COLNAMES,dtype='str')

        # replace empty strings with NaN
        rawdata = rawdata.apply(
            lambda x: x.str.strip()).replace('', np.nan)

        return rawdata


    def _clean_rawdata(self,rawdata):

        data = rawdata[self.KEEPCOLS].copy()
        for colname in list(data):

            if colname=='YYYYMMDD':
                data[colname] = pd.to_datetime(data[colname],
                    infer_datetime_format=True)
                data = data.set_index(
                    colname,verify_integrity=True)
                data.index.name='date'

            if colname=='RD':
                data[colname] = data[colname].astype(float)/10.

            if colname == 'SX':
                data[colname] = data[colname].replace(self.SNOWVALS)
                data[colname] = data[colname].astype(float)

            if colname == 'empty':
                data = data.drop(columns=[colname])

        return data


    def timeseries(self,var):
        """Return timeseries with data

        Parameters
        ----------
        varname : {'prc')
            variable to return
        """
 
        if var not in self.VARNAMES:
            msg = []
            warnings.warn((f'{var} is not a valid variable name. '
                f'parameter var must be in {self.VARNAMES} '
                f'by default precipitation data are returned.'))
            varname = 'prc'

        if var == 'prc':
            sr = self.data['RD']
            sr.name = 'prc'

        first = sr.first_valid_index()
        last = sr.sort_index(ascending=False).first_valid_index()
        return sr[first:last]


    @property
    def prc(self):
        """Return precipitation."""
        return self.timeseries('prc')


    @property
    def units(self):
        """Return table with definitions and units of variables"""

        tbl = pd.DataFrame({
            'variable' : ['prc','snow'],
            'datacol' : ['RH','EV24'],
            'unit' : ['mm/day','cm/day'], 
            })
        tbl = tbl.set_index('variable')
        return tbl




"""
Class BroGldXml reads XML with Groundwater Level Data from BRO XML 
file and also accepts result from a REST request.

"""
import os
import warnings
from pandas import Series,DataFrame
import pandas as pd
#import xml.etree.ElementTree as ET
from lxml.etree import _Element, _ElementTree
import lxml.etree as ET

class BroGldXml:
    """Read BRO Groundwater Level Data in XML tree format."""

    GLDPROPTAGS_XMLFILE = [
        ('dispatchTime','dispatchTime'),
        ('broIdGld','broId'),
        ('deliveryAccountableParty','deliveryAccountableParty'),
        ('qualityRegime','qualityRegime'),
        ('objectRegistrationTime','objectRegistrationTime'),
        ('registrationStatus','registrationStatus'),
        ('latestAdditionTime','latestAdditionTime'),
        ('corrected','corrected'),
        ('underReview','underReview'),
        ('deregistered','deregistered'),
        ('reregistered','reregistered'),
        ('researchFirstDate','ns4:researchFirstDate'),
        ('researchLastDate','ns4:researchLastDate'),
        ('broIdGmw','ns10:broId'),
        ('tubeNumber','ns10:tubeNumber'),
        ]

    GLDPROCESTAGS_XMLFILE = {
        'ObservationProcess':'ns5:ObservationProcess',
        'NamedValue':'ns6:NamedValue',
        'value':'ns6:value',
        }

    GLDOBSTAGS_XMLFILE = {
        'MeasurementTimeseries':'ns5:MeasurementTimeseries',
        'MeasurementTVP':'ns5:MeasurementTVP',
        'obstime':'ns5:time',
        'obsvalue':'ns5:value',
        'TVPMeasurementMetadata':'ns5:metadata/ns5:TVPMeasurementMetadata/ns5:qualifier/ns7:Category/ns7:value',
        }

    GLDOBSPROPSTAGS_XMLFILE = {
        'OM_Observation':'ns6:OM_Observation',
        'CI_RoleCode':'ns9:CI_RoleCode',
        'Date':'ns8:Date',
        }

    GLDPROPTAGS_RESTSERVICE = [
        ('dispatchTime','brocom:dispatchTime'),
        ('broIdGld','brocom:broId'),
        ('deliveryAccountableParty','brocom:deliveryAccountableParty'),
        ('qualityRegime','brocom:qualityRegime'),
        ('objectRegistrationTime','brocom:objectRegistrationTime'),
        ('registrationStatus','brocom:registrationStatus'),
        ('latestAdditionTime','brocom:latestAdditionTime'),
        ('corrected','brocom:corrected'),
        ('underReview','brocom:underReview'),
        ('deregistered','brocom:deregistered'),
        ('reregistered','brocom:reregistered'),
        ('researchFirstDate','ns11:researchFirstDate'),
        ('researchLastDate','ns11:researchLastDate'),
        ('broIdGmw','gldcommon:broId'),
        ('tubeNumber','gldcommon:tubeNumber'),
        ]

    GLDPROCESTAGS_RESTSERVICE = {
        'ObservationProcess':'waterml:ObservationProcess',
        'NamedValue':'om:NamedValue',
        'value':'om:value',
        }

    GLDOBSTAGS_RESTSERVICE = {
        'MeasurementTimeseries':'waterml:MeasurementTimeseries',
        'MeasurementTVP':'waterml:MeasurementTVP',
        'obstime':'waterml:time',
        'obsvalue':'waterml:value',
        'TVPMeasurementMetadata':'waterml:metadata/waterml:TVPMeasurementMetadata/waterml:qualifier/swe:Category/swe:value',
        }

    GLDOBSPROPSTAGS_RESTSERVICE = {
        'OM_Observation':'om:OM_Observation',
        'CI_RoleCode':'gmd:CI_RoleCode',
        'Date':'gco:Date',
        }


    def __init__(self,tree,xmlsource='file'):
        """
        Parameters
        ----------
        tree : ElementTree tree object
            Valid XML tree with groundwater level measurement data (GLD).
        xmlsource : {'file','rest'}, default 'file'
            Source of the XML tree (namespaces are different for 
            various sources).

        Example
        -------
        gld = BroGld(<elementtree>)"""

        self._tree = tree
        self._root = self._tree.getroot()
        self.NS = self._root.nsmap
        
        # define proper namespace tags for given source
        if xmlsource=='file':
            self.GLDPROPTAGS = self.GLDPROPTAGS_XMLFILE
            self.GLDPROCESTAGS = self.GLDPROCESTAGS_XMLFILE
            self.GLDOBSTAGS = self.GLDOBSTAGS_XMLFILE
            self.GLDOBSPROPSTAGS = self.GLDOBSPROPSTAGS_XMLFILE
        elif xmlsource=='rest':
            self.GLDPROPTAGS = self.GLDPROPTAGS_RESTSERVICE
            self.GLDPROCESTAGS = self.GLDPROCESTAGS_RESTSERVICE
            self.GLDOBSTAGS = self.GLDOBSTAGS_RESTSERVICE
            self.GLDOBSPROPSTAGS = self.GLDOBSPROPSTAGS_RESTSERVICE

    def __repr__(self):
        return self.gldprops['broIdGld']


    @classmethod
    def from_file(cls,xmlpath):
        """Read BRO GLD object from XML file.

        Parameters
        ----------
        xmlpath : str
            Valid filepath to XML file with groundwater level measurements.

        Returns
        -------
        BroGldXml object

        Example
        -------
        gld = BroGldXml.from_file(<valid xml filepath>)
        """

        if not os.path.isfile(xmlpath):
            raise ValueError(f'Invalid filepath: "{xmlpath}".')
        tree = ET.parse(xmlpath)
        return cls(tree,xmlsource='file')

    @classmethod
    def from_rest(cls,xmltree):
        """Parse BRO GLD tree returned from REST service.
        
        Parameters
        ----------
        xmltree : lxml tree
            Valid BRO GLD XML tree from BRO REST service.

        Returns
        -------
        BroGldXml object"""

        if isinstance(xmltree,_Element):
            # the BRO REST service returns an object of type _Element
            # not a tree
            xmltree = tree.getroottree()

        if not isinstance(xmltree,_ElementTree):
            raise ValueError(('Input is not a valid XML Element tree.'))

        tree = xmltree
        return cls(tree,xmlsource='rest')


    @property
    def gldprops(self):
        """Get level data properties."""
        gldprops = {}
        for key,tag in self.GLDPROPTAGS:
            node = self._root.find(f'.//{tag}',self.NS)
            if node is not None:
                gldprops[key]=node.text
            else:
                gldprops[key]=''
        return Series(gldprops,name='GldProperties')


    @property
    def procesprops(self):
        """Collect observation process parameters."""
        propslist = []
        
        # collect paramaters for all time series
        for node in self._root.iterfind(f'.//{self.GLDPROCESTAGS["ObservationProcess"]}',self.NS):
            props = {}
            
            # get time series id
            attdict = node.attrib
            tsid = [attdict[key] for key in attdict.keys()][0]
            tsid = tsid.split('_')[1]

            # get all parameters for time series
            for subnode in node.iterfind(f'.//{self.GLDPROCESTAGS["NamedValue"]}',self.NS):
                item = subnode.find(f'.//{self.GLDPROCESTAGS["value"]}',self.NS)
                prop = item.attrib['codeSpace'].split(':')[-1]
                props = {
                    'proces':tsid,
                    'parameter':prop,
                    'value':item.text,
                    }
                propslist.append(props)
                
        return DataFrame(propslist)

    @property
    def obs(self):
        """Get all observation data."""
        tslist = []
        for srnode in self._root.iterfind(f'.//{self.GLDOBSTAGS["MeasurementTimeseries"]}',self.NS):

            # get time series id
            attdict = srnode.attrib
            tsid = [attdict[key] for key in attdict.keys()][0]
            tsid = tsid.split('_')[1]

            # collect all time series
            ts = []
            for msnode in srnode.iterfind(f'.//{self.GLDOBSTAGS["MeasurementTVP"]}',self.NS):
                event = {
                    'time':msnode.find(f'.//{self.GLDOBSTAGS["obstime"]}',self.NS).text,
                    'value':msnode.find(f'.//{self.GLDOBSTAGS["obsvalue"]}',self.NS).text,
                    'quality':msnode.find(f'.//{self.GLDOBSTAGS["TVPMeasurementMetadata"]}',self.NS).text,
                    }
                ts.append(event)

            ts = DataFrame(ts)
            ts['timeseries'] = tsid
            tslist.append(ts)

        return pd.concat(tslist).sort_values(by='time').reset_index(drop=True)

    @property
    def obsprops(self):
        """Return observations metadata."""
        propslist = []
        for node in self._root.iterfind(F'.//{self.GLDOBSPROPSTAGS["OM_Observation"]}',self.NS):
            obsprops = {}
            attdict = node.attrib
            tsid = [attdict[key] for key in attdict.keys()][0]
            tsid = tsid.split('_')[1]
            obsprops['obsId']=tsid
            
            taglist = [
                ('obsRole',f'.//{self.GLDOBSPROPSTAGS["CI_RoleCode"]}'),
                ('dataStamp',f'.//{self.GLDOBSPROPSTAGS["Date"]}'),
                ]
            for key,tag in taglist:
                obsprops[key] = node.find(tag,self.NS).text
            propslist.append(Series(obsprops))

        return DataFrame(propslist)

    @property
    def heads(self):
        """Return time series with groundwater levels."""
        levels = self.obs['value'].astype(float).values
        datetimes = pd.to_datetime(self.obs['time'],infer_datetime_format=True)
        name = self.gldprops['broIdGld']
        heads = Series(data=levels,index=datetimes,name=name)
        if heads.index.has_duplicates:
            warnings.warn(f'Duplicate datetimes found in head series {name}.')
        return heads



""" 
Defines the KnmiLocations class that returns a list of available KNMI stations on the KNMI website and creates tables with station numbers 
and names.

Example
-------
>>> stations = aq.knmilocations()

"""

import os,os.path
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta
import warnings
import requests
import pkgutil
from io import StringIO
import warnings
import numpy as np
import pandas as pd
from pandas import Series, DataFrame
import pandas as pd
from json import JSONDecodeError

from ..geo.coordinate_conversion import CrdCon
##from ..data.knmi_data import knmi_prc_coords

def knmilocations(stntype='prc'):
    """ Return table with KNMI stations.
    
    Parameters
    ----------
    stntype : {'prc','wtr'}, default 'prc'
        type of knmi station

    Returns
    -------
    pd.DataFrame

    """

    knmi = KnmiDownload()

    if stntype not in ['prc','wtr',]:
        warnings.warn((f'Invalid stntype {stntype}. Default "prc" '
            f'will be used.'))
        stntype='prc'
    
    if stntype=='prc':
        tbl = knmi.prc_stns

    if stntype=='wtr':
        tbl = knmi.wtr_stns

    return tbl


class KnmiDownload:
    """Retrieve KNMI list of all available station numbers and names
    from KNMI website 

    Methods
    -------
    prc_stns(filepath=None)
        return list of manual rain gauche locations

    wtr_stns(filepath=None)
        return list of weather station locations

    Note
    ----
    Downloading lists of KNMI stations from the website takes time. 
    Recommended use is to download these data only once and save 
    results as csv for further use.

    """
    WEATHER_URL=r'https://www.daggegevens.knmi.nl/klimatologie/daggegevens'
    PRECIPITATION_URL=r'https://www.daggegevens.knmi.nl/klimatologie/monv/reeksen'
    BACKSHIFT = 2 #Backshift in months from today for retrieving one day of data

    WEATHER_HEADERLINE = '# STN         LON(east)   LAT(north)  ALT(m)      NAME'
    PRC_HEADERLINE = '# STN         NAME'

    #STN_COLS = ['stn_name','stn_type','xcr','ycr','lon','lat','alt']
    #STN_INDEXNAME = 'stn_nr'

    def __init__(self):

        # read list of precipitation station coordinates from 
        # local csv file within package
        #self.prc_crd = knmi_prc_coords()
        pass

    def __repr__(self):
        return self.__class__.__name__

    def _request_weather(self,par=None):
        """Request weather data and return server response."""

        if par is None:
            par = {
                'start':'20090817',
                'end':'20090817',
                'stns':'260',
                'vars':'RH:EV24',
                'fmt':'json'}

        # make actual request to knmi server
        self._response = requests.get(self.WEATHER_URL,params=par) 
        self._response_code = self._response.status_code
        self._response_url = self._response.url
        return self._response

    def _request_precipitation(self,par=None):
        """Request precipitation data and return server response."""

        if par is None:
            par = {
                'start':'20090817',
                'end':'20090817',
                'stns':'330',
                'fmt':'json'}

        # make actual request to knmi server
        self._response = requests.get(self.PRECIPITATION_URL,params=par)
        self._response_code = self._response.status_code
        self._response_url = self._response.url
        return self._response


    def download(self,kind='weather',start=None,end=None,stns=None,vars=None,
        result='data'):
        """Download KNMI weather station data.

        Parameters
        ----------
        kind : {'weather','prc'}, default 'weather'
            Measurement station type.
        start : str, optional (default yesterday)
            First day of download period (format as %Y%m%d).
        end : str, optional (default today)
            Last day of download period (format as %Y%m%d).
        stns : str or list of str (default all)
            Numbers of stations to download.
        vars : str, optional (default 'RH:EV24')
            Measured variables to download.
        result : {'data',text'}, default 'data'
            Output type.

        Returns
        -------
        Dataframe (data) or List of str (text)

        Notes
        -----
        With parameter output 'text' a list of strings is retured. This
        response is equal to the csv files that can be downloaded manually 
        from the KNMI website, including a large header with explanation
        of measured paramters. With parameter output 'data' a DataFrame
        with data is returned.
        """
        if kind not in ['weather','prc']:
            warnings.warn((f"Invalid measurement station type {kind}. "
                f"Default station type 'weather' is returned."))
            kind = 'weather'

        if result not in ['data','text']:
            warnings.warn((
                f"Invalid result parameter {result}. Data is returned."))
            result = 'json'

        # set request parameters
        if start is None:
            # publishing measurements can take asom time
            startday = (datetime.today()-relativedelta(
                months=self.BACKSHIFT))
            start = startday.strftime('%Y%m%d')
        if end is None:
            ##start_day = datetime.strptime(start, '%Y%m%d') #+timedelta(days=1)
            end = start

        if (stns is None) & (kind=='weather'):
            stns = '260' #'all'
        if (stns is None) & (kind=='prc'):
            stns = '330'

        fmt = 'çsv'
        if result=='data':
            fmt = 'json'

        # request data from server
        par = {'start':start,'end':end,'stns':stns,'fmt':fmt}
        if kind=='weather':
            if vars is None:
                vars = 'RH:EV24'
            par['vars']=vars
            response = self._request_weather(par=par)
        if kind=='prc':
            response = self._request_precipitation(par=par)

        # parse server response
        data = response.text #.splitlines()
        if fmt=='json':
            # parse result tp json
            try:
                data = response.json()
                data = DataFrame(data)
            except JSONDecodeError as err:
                raise JSONDecodeError((f'Response could not be serialised '
                    f'with request  {self._response_url}.'))
            if data.empty:
                warnings.warn((f'No data available for station {stns} '
                    f'inperiod {start} - {end}.'))

        return data


    def _findline(self,lines=None,tagline=None):
        """Find linenr of line in lines."""
        for i,line in enumerate(lines):
            if line.startswith(tagline):  
                start = i+1
                break
        if start is None:
            raise ValueError(f'Tagline not found: {tagline}.')
        return start

    def _find_first_non_numeric_line(self,lines=None,start=None):

        for j in range(start,len(lines)-1):
            line = lines[j]
            starts_with_number = line[1:9].strip().isnumeric()

            if not starts_with_number:
                end = j
                break
        return end

    @property
    def wtr_stns(self):
        """Return table of all available KNMI weather stations."""


        # download metadata for all weather stationa
        text = self.download(kind='weather',start=None,end=None,stns='all', #'260',
            vars=None,result='text')
        lines = text.splitlines()

        # find startline
        start = self._findline(lines=lines,tagline=self.WEATHER_HEADERLINE)

        # find endline
        end = self._find_first_non_numeric_line(lines=lines,start=start)

        # table stn numbers and metadata
        wht_stn=[]
        for i in range(start,end):

            parts = lines[i].split()
            name = ' '.join(parts[5:]) # 'De Bilt' was split...
            rec = {
                'stn_nr':int(parts[1]),
                'lon':parts[2],
                'lat':parts[3],
                'alt':parts[4],
                'stn_name':name,
                }
            wht_stn.append(rec)

        return DataFrame(wht_stn).set_index('stn_nr')

    @property
    def prc_stns(self):
        """ Return table of all available precipitation stations on KNMI site

        Notes
        -----
        Coordinates of precipitation stations are not available on the
        KNMI website.
        """

        text = self.download(kind='prc',stns='all',result='text')
        lines = text.splitlines()

        # find startline
        start = self._findline(lines=lines,tagline=self.PRC_HEADERLINE)

        # find endline
        end = self._find_first_non_numeric_line(
            lines=lines,start=start)

        # table stn numbers and metadata
        prc_stn=[]
        for i in range(start,end):

            parts = lines[i].split()
            name = ' '.join(parts[2:]) # 'De Bilt' was split...
            rec = {
                'stn_nr':int(parts[1]),
                'stn_name':name,
                }
            prc_stn.append(rec)

        return DataFrame(prc_stn).set_index('stn_nr')





import pytest
import numpy as np
from pandas import DataFrame
import pandas as pd
import matplotlib as mpl
from acequia import GwSeries,HeadsDif
import acequia as aq

@pytest.fixture
def gw1():
    dnpath = r'.\data\dinogws\B29A0848001_1.csv'
    return GwSeries.from_dinogws(dnpath)


@pytest.fixture
def gw2():
    dnpath = r'.\data\dinogws\B29A0848002_1.csv'
    return GwSeries.from_dinogws(dnpath)

@pytest.fixture
def gw3():
    dnpath = r'.\data\dinogws\B29A0848003_1.csv'
    return GwSeries.from_dinogws(dnpath)

@pytest.fixture
def hdf(gw1,gw2):
    locname = gw1.locname()
    refcol = gw1.name()
    return HeadsDif.from_series(heads=[gw1,gw2],locname=locname,
        refcol=refcol)

   
def test_from_series(gw1,gw2):
    """Test classmethod from_series with different input."""

    # test with gwseries
    hdf = HeadsDif.from_series(heads=[gw1,gw2],locname=None,refcol=None)
    assert isinstance(hdf.heads,DataFrame)
    assert not hdf.heads.empty

    # test with pandas series
    sr1 = gw1.heads()
    sr2 = gw2.heads()
    hdf = HeadsDif.from_series(heads=[sr1,sr2],locname=None,refcol=None)
    assert isinstance(hdf.heads,DataFrame)
    assert not hdf.heads.empty
    
    # test UserWarning for invalid refcol
    with pytest.warns(UserWarning) as record:
        hdf = HeadsDif.from_series(heads=[gw1,gw2],locname=None,
            refcol='onzinkolom')
        assert len(record)==1


def test_repr(hdf):
    assert isinstance(hdf.__repr__(),str)


def test_get_difference(hdf):

    assert isinstance(hdf.get_difference(),DataFrame)
    assert not hdf.get_difference().empty
    
    # test UserWarning for invalid refcol
    with pytest.warns(UserWarning) as record:
         hdf.get_difference(refcol='onzinkolom')
         assert len(record)==1

def test_relative_heads(hdf):
    relheads = hdf.get_relative_heads()
    assert isinstance(relheads,DataFrame)
    assert not relheads.empty

def test_get_seasons(hdf):

    seasons = hdf.get_seasons()
    assert len(seasons)!=0
    
    seasons = hdf.get_seasons(period='seasons')
    assert len(seasons)!=0

    seasons = hdf.get_seasons(period='biannual')
    assert len(seasons)!=0

    seasons = hdf.get_seasons(dates=hdf.heads)
    assert len(seasons)!=0

    seasons = hdf.get_seasons(dates=hdf.heads.index)
    assert len(seasons)!=0

def test_get_difference_by_season(hdf):

    df = hdf.get_difference_by_season()
    assert isinstance(df,DataFrame)

    with pytest.warns(UserWarning) as record:
        df = hdf.get_difference_by_season(period='foute invoer')
        assert len(record)==1

    df = hdf.get_difference_by_season(period=hdf.PERIOD_DEFAULT)
    assert isinstance(df,DataFrame)

    for season in hdf.PERIOD_NAMES:
        df = hdf.get_difference_by_season(period=season)
        assert isinstance(df,DataFrame)
        assert not df.empty

def test_plot_time(gw1,gw2,gw3):

    hdf = HeadsDif.from_series(heads=[gw1,gw2,gw3])
    axlist = hdf.plot_time()
    assert isinstance(axlist,np.ndarray)
    for ax in axlist.flatten():
        #assert isinstance(ax,mpl.axes._subplots.Axes)
        assert isinstance(ax,mpl.axes.Axes)

def test_plot_head(gw1,gw2,gw3):

    hdf = HeadsDif.from_series(heads=[gw1,gw2,gw3])
    axlist = hdf.plot_head()
    assert isinstance(axlist,np.ndarray)
    for ax in axlist.flatten():
        #assert isinstance(ax,mpl.axes._subplots.Axes)
        assert isinstance(ax,mpl.axes.Axes)

def test_plot_freq(gw1,gw2,gw3):

    hdf = HeadsDif.from_series(heads=[gw1,gw2,gw3])
    axlist = hdf.plot_freq()
    assert isinstance(axlist,np.ndarray)
    for ax in axlist.flatten():
        #assert isinstance(ax,mpl.axes._subplots.Axes)
        assert isinstance(ax,mpl.axes.Axes)


""" 
This module contains the object GwLocs for managing data from multiple 
well locations.
"""

import os
import warnings
from pandas import Series, DataFrame
import pandas as pd
import numpy as np

from .gwseries import GwSeries
from .read.filedirtools import listdir

class GwLocs:
    """Manage multiple groundwater heads series from one well location



    Methods
    -------
    sourcetable()
        return list of location and heads series names

    gwseries(self,loc=None)
        return list of GwSeries objects

    Examples
    --------
    Create GwLocs object:
    >>>locs = GwLocs(filedir=<jsondir>)
    Return table of locations and series:
    >>>tbl = locs.loctable()
    Return al series for location B16D0037:
    >>>gws = GwLocs.gwseries(loc='B16D0037')
    Return list of GwSeries objects for locations in list:
    >>>names = ['B16D0037','B27G0237',['B28D1635','B28B1388'],'B28B1389']
    >>>for loc in GwLocs(filedir=<jsondir>,groups=names):
            print(f'{names[i]} group size is {len(gws)}')
    Explicitly iterate over locations:
    >>>locs = GwLocs(filedir=<jsondir>,groups=names)
    >>>for i in range(len(locs)):
            gws = next(locs)
            print(f'{names[i]} group size is {len(gws)}')
    """

    def __init__(self,filedir=None,pathlist=None,filetype=None,
        groups=None):
        """
        Parameters
        ----------
        filedir : str
            directory with source files

        pathlist : list, optional
            list of sourcefile names (without path)

        filetype : ['.csv','.json'], optional
            source file type 

        groups : list, optional
            list of location names, sublists are allowed (see examples)
        """

        message = ('Class GwList will be deprecated in a future version.'
            'Use GwCollection class instead.')
        warnings.warn(message, DeprecationWarning, stacklevel=2)

        self._filedir = filedir
        self._pathlist = pathlist
        self._filetype = filetype
        self._loclist = groups

        if not isinstance(self._filedir,str):
            msg = f'filedir must be a string, not {self._filedir}'
            raise ValueError(msg)

        if not os.path.isdir(self._filedir):
            msg = f'{self._filedir} is not a valid directory'
            raise ValueError(msg)

        if self._pathlist is None:
            self._pathlist = listdir(dirpath=self._filedir)

        if self._filetype is None:
            self._filetype = self._infer_filetype()

        if self._filetype in ['json','csv']:
            self._filetype = '.'+self._filetype

        if self._filetype not in ['.json','.csv']:
            msg = f'Unsupported filetype {filetype}.'
            raise ValueError(msg)

        ##self._srcfiles = self._srctbl()
        ##self._locfiles = self.list_locfiles()
        self._create_iterator()


    def __repr__(self):
        nlocs = len(self._srctable().groupby(by='loc').size())
        return (f'GwLocs(nlocs={nlocs})')


    def _infer_filetype(self):
        """Infer filetype from file extensin count in <filedir>"""
        filelist = [os.path.basename(path) for path in self._pathlist]
        extlist = [os.path.splitext(filename)[1] for filename in filelist]
        sr = Series(extlist).value_counts()
        return sr.index[0]


    def _srctable(self):
        """Return table of all valid source files"""

        if self._filetype=='.json':
            filepaths = [x for x in self._pathlist]
            fnames = [os.path.basename(path) for path in self._pathlist]
            ftypes = [os.path.splitext(name)[1] for name in fnames]
            series = [os.path.splitext(name)[0] for name in fnames]
            locations = [x.split('_')[0] for x in series]
            filters = [x.split('_')[1] for x in series]

        if self._filetype=='.csv':
            # dino zipfiles come with two files for each filter
            # only files ending with _1.csv are needed
            # filename: B01C0008001_1.csv
            filepaths = [x for x in self._pathlist if x.endswith('_1.csv')]
            fnames = [os.path.basename(x) for x in filepaths]
            ftypes = [os.path.splitext(name)[1] for name in fnames]
            locations = [x[:8] for x in fnames]
            filters = [x[8:11].lstrip('0') for x in fnames]
            series = [x[0]+'_'+x[1] for x in list(zip(locations,filters))]

        table = DataFrame(data={
                'loc':locations,
                'fil':filters,
                'filename':fnames,
                'filetype':ftypes,
                'filepaths':filepaths,
                },index=series)
        table.index.name = 'series'
        return table


    """
    def list_sourcefiles(self,filedir=None):
        """ """Return table with list of series names and well location 
        names based on sourcefile names in directory filedir

        Parameters
        ----------
        filedir : str, optional
            directory with sourcefile names

        Returns
        -------
        pd.DataFrame

        Notes
        -----
        Directory <filedir> should contain source files with filenames
        composed of <location name>_<filter>.<extension>, like in 
        'B28H0025_1.csv' or 'Mylocation_Myfilter.json'.

        """ """

        if filedir is None:
            return self._srcfiles

        if not os.path.isdir(filedir):
            msg = f'{filedir} must be a valid directory'
            raise ValueError(msg)

        return self._srctable(filedir)
    """

    def filetable(self):
        """Return table of source files for all series"""

        srctbl = self._srctable()

        if self._loclist is not None:

            def flatten_nested_list(nested_list):
                # stackoverflow.com/questions/10823877/
                for i in nested_list:
                    if isinstance(i, (list,tuple)):
                        for j in flatten_nested_list(i):
                            yield j
                    else:
                        yield i

            flatlist = flatten_nested_list(self._loclist)
            mask = srctbl['loc'].isin(flatlist)
            srctbl = srctbl[mask]

        if srctbl.empty:
            msg = f'list_locfiles returns empty dataframe.'
            warnings.warn(msg)

        return srctbl


    def _create_iterator(self):
        """Create location series groups iterator """
        self._itercount = 0
        if self._loclist is None:
            ##grp = self._srcfiles.groupby(by='loc')
            ##grp = self._locfiles.groupby(by='loc')
            grp = self.filetable().groupby(by='loc')
            self._loclist = list(grp.groups.keys())
        else:
            self._loclist = self._loclist
            #todo: check if groupnames are in table


    def gwseries(self,loc=None):
        """Return list of GwSeries objects with location name <loc>

        Parameters
        ----------
        loc : str
            name of location

        Return
        ------
        list of GwSeries objects

        Notes
        -----
        When parameter loc is a list of locations names, GwSeries objects
        for all these locations will be returned.
        """

        if isinstance(loc,str):
            loc = [loc]

        mask = self._srctable()['loc'].isin(loc)
        gwlist = []
        for i,(srname,row) in enumerate(self._srctable()[mask].iterrows()):

            filepath = os.path.join(self._filedir,row['filename'])
            if self._filetype=='.json':
                gw = GwSeries.from_json(filepath)
            if self._filetype=='.csv':
                gw = GwSeries.from_dinogws(filepath)

            #if i==0:
            #    gwlist = []
            gwlist.append(gw)

        return gwlist


    def __iter__(self):
        return self


    def __len__(self):
        return len(self._loclist)


    def __next__(self):

        if self._itercount >= len(self):
            raise StopIteration

        loc = self._loclist[self._itercount]
        mygwlist = self.gwseries(loc=loc)

        self._itercount+=1
        return mygwlist

